@@ -23,6 +23,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <unistd.h>
+#include <assert.h>
 
 #include <tss2/tss2_tpm2_types.h>
 
@@ -31,6 +32,7 @@
 #include "runtime.h"
 #include "digest.h"
 #include "util.h"
+#include "uapi.h"
 
 #define TPM_EVENT_LOG_MAX_ALGOS		64
 
@@ -404,14 +406,10 @@ tpm_event_type_to_string(unsigned int event_type)
 }
 
 const tpm_evdigest_t *
-tpm_event_get_digest(const tpm_event_t *ev, const char *algo_name)
+tpm_event_get_digest(const tpm_event_t *ev, const tpm_algo_info_t *algo_info)
 {
-	const tpm_algo_info_t *algo_info;
 	unsigned int i;
 
-	if ((algo_info = digest_by_name(algo_name)) < 0)
-		fatal("Unknown algo name \"%s\"\n", algo_name);
-
 	for (i = 0; i < ev->pcr_count; ++i) {
 		const tpm_evdigest_t *md = &ev->pcr_values[i];
 
@@ -551,7 +549,7 @@ tpm_event_decode_uuid(const unsigned char *data)
 }
 
 /*
- * Handle IPL events, which grub2 uses to hide its stuff in
+ * Handle IPL events, which grub2 and sd-boot uses to hide its stuff in
  */
 static void
 __tpm_event_grub_file_destroy(tpm_parsed_event_t *parsed)
@@ -764,6 +762,156 @@ __tpm_event_shim_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed, const
 	return true;
 }
 
+static void
+__tpm_event_systemd_destroy(tpm_parsed_event_t *parsed)
+{
+	drop_string(&parsed->systemd_event.string);
+}
+
+static const char *
+__tpm_event_systemd_describe(const tpm_parsed_event_t *parsed)
+{
+	static char buffer[1024];
+	char data[768];
+	unsigned int len;
+
+	/* It is in UTF16, and also include two '\0' at the end */
+	len = parsed->systemd_event.len >> 1;
+	if (len > sizeof(data))
+		len = sizeof(data);
+	__convert_from_utf16le(parsed->systemd_event.string, parsed->systemd_event.len, data, len);
+	data[len] = '\0';
+
+	snprintf(buffer, sizeof(buffer), "systemd boot event %s", data);
+	return buffer;
+}
+
+static const tpm_evdigest_t *
+__tpm_event_systemd_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *parsed, tpm_event_log_rehash_ctx_t *ctx)
+{
+	const uapi_boot_entry_t *boot_entry = ctx->boot_entry;
+	char initrd[2048];
+	char initrd_utf16[4096];
+	unsigned int len;
+
+	/* If no --next-kernel option was given, do not rehash anything */
+	if (boot_entry == NULL)
+		return tpm_event_get_digest(ev, ctx->algo);
+
+	if (!boot_entry->image_path) {
+		error("Unable to identify the next kernel\n");
+		return NULL;
+	}
+
+	debug("Next boot entry expected from: %s %s\n", boot_entry->title, boot_entry->version? : "");
+	snprintf(initrd, sizeof(initrd), "initrd=%s %s",
+			path_unix2dos(boot_entry->initrd_path),
+			boot_entry->options? : "");
+
+	len = (strlen(initrd) + 1) << 1;
+	assert(len <= sizeof(initrd_utf16));
+	__convert_to_utf16le(initrd, strlen(initrd) + 1, initrd_utf16, len);
+
+	return digest_compute(ctx->algo, initrd_utf16, len);
+}
+
+/*
+ * This event holds stuff like
+ *  initrd = ....
+ */
+static bool
+__tpm_event_systemd_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed, const char *value, unsigned int len)
+{
+	struct systemd_event *evspec = &parsed->systemd_event;
+
+	evspec->len = len;
+	evspec->string = malloc(len);
+	memcpy(evspec->string, value, len);
+
+	parsed->event_subtype = SYSTEMD_EVENT_VARIABLE;
+	parsed->destroy = __tpm_event_systemd_destroy;
+	parsed->rehash = __tpm_event_systemd_rehash;
+	parsed->describe = __tpm_event_systemd_describe;
+
+	return true;
+}
+
+
+static void
+__tpm_event_tag_destroy(tpm_parsed_event_t *parsed)
+{
+}
+
+static const char *
+__tpm_event_tag_options_describe(const tpm_parsed_event_t *parsed)
+{
+	return "Kernel command line (measured by the kernel)";
+}
+
+static const tpm_evdigest_t *
+__tpm_event_tag_options_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *parsed, tpm_event_log_rehash_ctx_t *ctx)
+{
+	return __tpm_event_systemd_rehash(ev, parsed, ctx);
+}
+
+static const char *
+__tpm_event_tag_initrd_describe(const tpm_parsed_event_t *parsed)
+{
+	return "initrd (measured by the kernel)";
+}
+
+static const tpm_evdigest_t *
+__tpm_event_tag_initrd_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *parsed, tpm_event_log_rehash_ctx_t *ctx)
+{
+	const uapi_boot_entry_t *boot_entry = ctx->boot_entry;
+
+	/* If no --next-kernel option was given, do not rehash anything */
+	if (boot_entry == NULL)
+		return tpm_event_get_digest(ev, ctx->algo);
+
+	if (!boot_entry->initrd_path) {
+		/* Can this happen eg when going from a split kernel to a unified kernel? */
+		error("Unable to identify the next initrd\n");
+		return NULL;
+	}
+
+	debug("Next boot entry expected from: %s %s\n", boot_entry->title, boot_entry->version? : "");
+	return runtime_digest_efi_file(ctx->algo, boot_entry->initrd_path);
+}
+
+/*
+ * Generated by the kernel (PCR#9), to measure the cmdline and initrd
+ */
+static bool
+__tpm_event_parse_tag(tpm_event_t *ev, tpm_parsed_event_t *parsed, buffer_t *bp)
+{
+	struct tag_event *evspec = &parsed->tag_event;
+
+	if (!buffer_get_u32le(bp, &evspec->event_id))
+		return false;
+
+	if (!buffer_get_u32le(bp, &(evspec->event_data_len)))
+		return false;
+
+	if (evspec->event_data_len > sizeof(evspec->event_data))
+		return false;
+
+	if (!buffer_get(bp, evspec->event_data, evspec->event_data_len))
+		return false;
+
+	parsed->destroy = __tpm_event_tag_destroy;
+	if (evspec->event_id == LOAD_OPTIONS_EVENT_TAG_ID) {
+		parsed->rehash = __tpm_event_tag_options_rehash;
+		parsed->describe = __tpm_event_tag_options_describe;
+	} else
+	if (evspec->event_id == INITRD_EVENT_TAG_ID) {
+		parsed->rehash = __tpm_event_tag_initrd_rehash;
+		parsed->describe = __tpm_event_tag_initrd_describe;
+	} else
+		return false;
+
+	return true;
+}
 
 static bool
 __tpm_event_parse_ipl(tpm_event_t *ev, tpm_parsed_event_t *parsed, buffer_t *bp)
@@ -788,6 +936,9 @@ __tpm_event_parse_ipl(tpm_event_t *ev, tpm_parsed_event_t *parsed, buffer_t *bp)
 	if (ev->pcr_index == 9)
 		return __tpm_event_grub_file_event_parse(ev, parsed, value);
 
+	if (ev->pcr_index == 12)
+		return __tpm_event_systemd_event_parse(ev, parsed, value, len);
+
 	if (ev->pcr_index == 14)
 		return __tpm_event_shim_event_parse(ev, parsed, value);
 
@@ -802,6 +953,9 @@ __tpm_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed, tpm_event_log_sca
 	buffer_init_read(&buf, ev->event_data, ev->event_size);
 
 	switch (ev->event_type) {
+	case TPM2_EVENT_EVENT_TAG:
+		return __tpm_event_parse_tag(ev, parsed, &buf);
+
 	case TPM2_EVENT_IPL:
 		return __tpm_event_parse_ipl(ev, parsed, &buf);
 
