@@ -32,7 +32,9 @@
 #include "pcr.h"
 #include "digest.h"
 #include "rsa.h"
+#include "store.h"
 #include "testcase.h"
+#include "sd-boot.h"
 
 enum {
 	ACTION_NONE,
@@ -43,6 +45,7 @@ enum {
 	ACTION_UNSEAL,
 	ACTION_SIGN,
 	ACTION_SELFTEST,
+	ACTION_RSATEST,
 };
 
 enum {
@@ -56,6 +59,7 @@ struct predictor {
 	const char *		initial_source;
 
 	const char *		tpm_event_log_path;
+	const char *		boot_entry_id;
 
 	const char *		algo;
 	const tpm_algo_info_t *	algo_info;
@@ -87,12 +91,16 @@ enum {
 	OPT_RSA_PRIVATE_KEY,
 	OPT_RSA_PUBLIC_KEY,
 	OPT_RSA_GENERATE_KEY,
+	OPT_RSA_BITS,
 	OPT_INPUT,
 	OPT_OUTPUT,
 	OPT_AUTHORIZED_POLICY,
 	OPT_PCR_POLICY,
 	OPT_KEY_FORMAT,
 	OPT_POLICY_NAME,
+	OPT_POLICY_FORMAT,
+	OPT_TARGET_PLATFORM,
+	OPT_BOOT_ENTRY,
 };
 
 static struct option options[] = {
@@ -109,18 +117,23 @@ static struct option options[] = {
 	{ "before",		no_argument,		0,	OPT_BEFORE },
 	{ "verify",		required_argument,	0,	OPT_VERIFY },
 	{ "use-pesign",		no_argument,		0,	OPT_USE_PESIGN },
+	{ "boot-entry",		required_argument,	0,	OPT_BOOT_ENTRY },
 	{ "create-testcase",	required_argument,	0,	OPT_CREATE_TESTCASE },
 	{ "replay-testcase",	required_argument,	0,	OPT_REPLAY_TESTCASE },
 
 	{ "private-key",	required_argument,	0,	OPT_RSA_PRIVATE_KEY },
 	{ "public-key",		required_argument,	0,	OPT_RSA_PUBLIC_KEY },
 	{ "rsa-generate-key",	no_argument,		0,	OPT_RSA_GENERATE_KEY },
+	{ "rsa-bits",		required_argument,	0,	OPT_RSA_BITS },
 	{ "input",		required_argument,	0,	OPT_INPUT },
 	{ "output",		required_argument,	0,	OPT_OUTPUT },
 	{ "authorized-policy",	required_argument,	0,	OPT_AUTHORIZED_POLICY },
 	{ "pcr-policy",		required_argument,	0,	OPT_PCR_POLICY },
 	{ "key-format",		required_argument,	0,	OPT_KEY_FORMAT },
 	{ "policy-name",	required_argument,	0,	OPT_POLICY_NAME },
+	{ "policy-format",	required_argument,	0,	OPT_POLICY_FORMAT },
+	{ "target-platform",	required_argument,	0,	OPT_TARGET_PLATFORM },
+	{ "next-kernel",	required_argument,	0,	OPT_BOOT_ENTRY },
 
 	{ NULL }
 };
@@ -242,7 +255,8 @@ predictor_load_eventlog(struct predictor *pred)
 static struct predictor *
 predictor_new(const tpm_pcr_selection_t *pcr_selection, const char *source,
 		const char *tpm_eventlog_path,
-		const char *output_format)
+		const char *output_format,
+		const char *boot_entry_id)
 {
 	struct predictor *pred;
 
@@ -252,6 +266,7 @@ predictor_new(const tpm_pcr_selection_t *pcr_selection, const char *source,
 	pred = calloc(1, sizeof(*pred));
 	pred->pcr_mask = pcr_selection->pcr_mask;
 	pred->initial_source = source;
+	pred->boot_entry_id = boot_entry_id;
 
 	pred->algo = pcr_selection->algo_info->openssl_name;
 	pred->algo_info = pcr_selection->algo_info;
@@ -554,6 +569,9 @@ predictor_get_event_strategy(unsigned int event_type)
 		/* IPL: used by grub2 for PCR 8 and PCR9 */
 		TPM2_EVENT_IPL,
 
+		/* EVENT_TAG: used by the kernel for PCR9, to measure the cmdline and initrd */
+		TPM2_EVENT_EVENT_TAG,
+
 		/*
 		 * EFI_GPT_EVENT: used in updates of PCR5, seems to be a hash of several GPT headers.
 		 *	We should probably rebuild in case someone changed the partitioning.
@@ -640,6 +658,14 @@ predictor_update_eventlog(struct predictor *pred)
 	tpm_event_log_rehash_ctx_init(&rehash_ctx, pred->algo_info);
 	rehash_ctx.use_pesign = opt_use_pesign;
 
+	/* The argument given to --next-kernel will be either "auto" or the
+	 * systemd ID of the next kernel entry to be booted.
+	 * FIXME: we should probably hide this behind a target_platform function.
+	 */
+	if (pred->boot_entry_id != NULL
+	 && !(rehash_ctx.boot_entry = sdb_identify_boot_entry(pred->boot_entry_id)))
+		fatal("unable to identify next kernel \"%s\"\n", pred->boot_entry_id);
+
 	for (ev = pred->event_log; ev; ev = ev->next) {
 		tpm_evdigest_t *pcr;
 		bool stop = false;
@@ -659,7 +685,7 @@ predictor_update_eventlog(struct predictor *pred)
 			debug("\n");
 			__tpm_event_print(ev, debug);
 
-			if (!(old_digest = tpm_event_get_digest(ev, pred->algo)))
+			if (!(old_digest = tpm_event_get_digest(ev, pred->algo_info)))
 				fatal("Event log lacks a hash for digest algorithm %s\n", pred->algo);
 
 			if (false) {
@@ -942,6 +968,7 @@ get_action_argument(int argc, char **argv)
 		{ "unseal-secret",		ACTION_UNSEAL	},
 		{ "sign",			ACTION_SIGN	},
 		{ "self-test",			ACTION_SELFTEST	},
+		{ "rsa-test",			ACTION_RSATEST	},
 
 		{ NULL, 0 },
 	};
@@ -989,7 +1016,7 @@ main(int argc, char **argv)
 {
 	struct predictor *pred;
 	int action = ACTION_NONE;
-	tpm_pcr_selection_t *pcr_selection;
+	tpm_pcr_selection_t *pcr_selection = NULL;
 	char *opt_from = NULL;
 	char *opt_algo = NULL;
 	char *opt_output_format = NULL;
@@ -1003,12 +1030,16 @@ main(int argc, char **argv)
 	char *opt_output = NULL;
 	char *opt_authorized_policy = NULL;
 	char *opt_pcr_policy = NULL;
-	char *opt_rsa_private_key = NULL;
-	char *opt_rsa_public_key = NULL;
+	stored_key_t *opt_rsa_private_key = NULL;
+	stored_key_t *opt_rsa_public_key = NULL;
 	bool opt_rsa_generate = false;
-	char *opt_key_format = NULL;
+	char *opt_rsa_bits = NULL;
 	char *opt_policy_name = NULL;
-	bool tpm2key_fmt = false;
+	char *opt_target_platform = NULL;
+	char *opt_boot_entry = NULL;
+	const target_platform_t *target;
+	unsigned int action_flags = 0;
+	unsigned int rsa_bits = 2048;
 	int c, exit_code = 0;
 
 	while ((c = getopt_long(argc, argv, "dhA:CF:LSZ", options, NULL)) != EOF) {
@@ -1040,6 +1071,9 @@ main(int argc, char **argv)
 		case OPT_USE_PESIGN:
 			opt_use_pesign = 1;
 			break;
+		case OPT_BOOT_ENTRY:
+			opt_boot_entry = optarg;
+			break;
 		case OPT_STOP_EVENT:
 			opt_stop_event = optarg;
 			break;
@@ -1062,14 +1096,19 @@ main(int argc, char **argv)
 			opt_replay_testcase = optarg;
 			break;
 		case OPT_RSA_PRIVATE_KEY:
-			opt_rsa_private_key = optarg;
+			/* The private key file uses PEM format by default */
+			opt_rsa_private_key = stored_key_new_private(STORED_KEY_FMT_PEM, optarg);
 			break;
 		case OPT_RSA_PUBLIC_KEY:
-			opt_rsa_public_key = optarg;
+			/* The public key file uses native TPM format by default */
+			opt_rsa_public_key = stored_key_new_private(STORED_KEY_FMT_NATIVE, optarg);
 			break;
 		case OPT_RSA_GENERATE_KEY:
 			opt_rsa_generate = true;
 			break;
+		case OPT_RSA_BITS:
+			opt_rsa_bits = optarg;
+			break;
 		case OPT_INPUT:
 			opt_input = optarg;
 			break;
@@ -1083,11 +1122,19 @@ main(int argc, char **argv)
 			opt_pcr_policy = optarg;
 			break;
 		case OPT_KEY_FORMAT:
-			opt_key_format = optarg;
+			warning("Detected --key-format option; please use --target-platform instead\n");
+			opt_target_platform = optarg;
 			break;
 		case OPT_POLICY_NAME:
 			opt_policy_name = optarg;
 			break;
+		case OPT_POLICY_FORMAT:
+			warning("Detected --policy-format option; please use --target-platform instead\n");
+			opt_target_platform = optarg;
+			break;
+		case OPT_TARGET_PLATFORM:
+			opt_target_platform = optarg;
+			break;
 		case 'h':
 			usage(0, NULL);
 		default:
@@ -1106,13 +1153,23 @@ main(int argc, char **argv)
 	if (opt_create_testcase)
 		runtime_record_testcase(testcase_alloc(opt_create_testcase));
 
-	if (!opt_key_format || !strcasecmp(opt_key_format, "raw"))
-		tpm2key_fmt = false;
-	else
-	if (!strcasecmp(opt_key_format, "tpm2.0"))
-		tpm2key_fmt = true;
-	else
-		fatal("Unsupported key format \"%s\"\n", opt_key_format);
+	if (opt_rsa_bits) {
+		if (strcmp(opt_rsa_bits, "2048") == 0)
+			rsa_bits = 2048;
+		else
+		if (strcmp(opt_rsa_bits, "3072") == 0)
+			rsa_bits = 3072;
+		else
+		if (strcmp(opt_rsa_bits, "4096") == 0)
+			rsa_bits = 4096;
+		else
+			fatal("Unsupported RSA bits: %s\n", opt_rsa_bits);
+	}
+
+	if (opt_target_platform == NULL)
+		opt_target_platform = "tpm2.0";
+	if ((target = pcr_get_target_platform(opt_target_platform)) == NULL)
+		fatal("Unsupported target platform %s\n", opt_target_platform);
 
 	/* Validate options */
 	switch (action) {
@@ -1124,8 +1181,8 @@ main(int argc, char **argv)
 	case ACTION_STORE_PUBLIC_KEY:
 		if (opt_rsa_private_key == NULL)
 			usage(1, "You need to specify the RSA secret key using --private-key option\n");
-		if (opt_rsa_public_key == NULL)
-			opt_rsa_public_key = opt_output;
+		if (opt_rsa_public_key == NULL && opt_output)
+			opt_rsa_public_key = stored_key_new_public(STORED_KEY_FMT_NATIVE, opt_output);
 		end_arguments(argc, argv);
 		break;
 
@@ -1147,26 +1204,29 @@ main(int argc, char **argv)
 		break;
 
 	case ACTION_UNSEAL:
-		if (tpm2key_fmt) {
-			end_arguments(argc, argv);
-			break;
+		action_flags = target_platform_unseal_flags(target);
+		if (action_flags & PLATFORM_OPTIONAL_PCR_POLICY) {
+			if (opt_rsa_public_key == NULL && opt_pcr_policy)
+				usage(1, "You need to specify the --public-key option when unsealing using an authorized policy\n");
+			if (opt_pcr_policy == NULL && opt_rsa_public_key)
+				usage(1, "You need to specify the --pcr-policy option when unsealing using an authorized policy\n");
 		}
 
-		if (opt_rsa_public_key == NULL && opt_pcr_policy)
-			usage(1, "You need to specify the --public-key option when unsealing using an authorized policy\n");
-		if (opt_pcr_policy == NULL && opt_rsa_public_key)
-			usage(1, "You need to specify the --pcr-policy option when unsealing using an authorized policy\n");
-		pcr_selection = get_pcr_selection_argument(argc, argv, opt_algo);
+		if ((action_flags & PLATFORM_NEED_INPUT_FILE) && !opt_input)
+			usage(1, "You need to specify an input file via --input when unsealing a secret");
+		if ((action_flags & PLATFORM_NEED_OUTPUT_FILE) && !opt_output)
+			usage(1, "You need to specify an output file via --output when unsealing a secret");
+		if (action_flags & PLATFORM_NEED_PCR_SELECTION)
+			pcr_selection = get_pcr_selection_argument(argc, argv, opt_algo);
 		end_arguments(argc, argv);
 		break;
 
 	case ACTION_SIGN:
 		if (opt_rsa_private_key == NULL)
 			usage(1, "You need to specify the --private-key option when signing a policy\n");
-		if (tpm2key_fmt) {
-			if (opt_input == NULL)
-				usage(1, "You need to specify the --input option when signing a policy into a TPM 2.0 Key file\n");
-		}
+		if (opt_output == NULL)
+			usage(1, "You need to specify the --output option when signing a policy\n");
+
 		pcr_selection = get_pcr_selection_argument(argc, argv, opt_algo);
 		end_arguments(argc, argv);
 		break;
@@ -1174,8 +1234,43 @@ main(int argc, char **argv)
 	case ACTION_SELFTEST:
 		end_arguments(argc, argv);
 		break;
+
+	case ACTION_RSATEST:
+		end_arguments(argc, argv);
+		break;
+
+	default:
+		fatal("Action %u not implemented", action);
 	}
 
+	/* If we're asked to generate a new RSA key, do so.  This
+	 * doesn't add anything beyond what "openssl genrsa" would do,
+	 * except it saves you from installing the openssl tool suite
+	 * if you don't want it. */
+	if (opt_rsa_generate && (action == ACTION_CREATE_AUTH_POLICY ||
+				 action == ACTION_STORE_PUBLIC_KEY ||
+				 action == ACTION_SIGN)) {
+		tpm_rsa_key_t *key;
+
+		infomsg("Generating new RSA key\n");
+		if (!(key = tpm_rsa_generate(rsa_bits)))
+			return 1;
+		if (!stored_key_write_rsa_private(opt_rsa_private_key, key))
+			return 1;
+	}
+
+	if (action == ACTION_RSATEST) {
+		if (tpm_rsa_bits_test(rsa_bits)) {
+			infomsg("RSA %u supported\n", rsa_bits);
+			return 0;
+		} else {
+			infomsg("RSA %u unsupported\n", rsa_bits);
+			return 1;
+		}
+	}
+
+	set_srk_rsa_bits (rsa_bits);
+
 	if (action == ACTION_SELFTEST) {
 		if (!tpm_selftest(true))
 			return 1;
@@ -1184,20 +1279,6 @@ main(int argc, char **argv)
 	}
 
 	if (action == ACTION_CREATE_AUTH_POLICY) {
-		/* If we're asked to generate a new RSA key, do so.
-		 * This doesn't add anything beyond what "openssl genrsa" would do,
-		 * except it saves you from installing the openssl tool suite if you
-		 * don't want it. */
-		if (opt_rsa_generate) {
-			tpm_rsa_key_t *key;
-
-			infomsg("Generating new RSA key\n");
-			if (!(key = tpm_rsa_generate(2048)))
-				return 1;
-			if (!tpm_rsa_key_write_private(opt_rsa_private_key, key))
-				return 1;
-		}
-
 		if (!pcr_authorized_policy_create(pcr_selection, opt_rsa_private_key, opt_authorized_policy))
 			return 1;
 
@@ -1213,26 +1294,15 @@ main(int argc, char **argv)
 	/* When sealing a secret against an authorized policy, there's no need to
 	 * mess around with PCR values. That's the beauty of it... */
 	if (action == ACTION_SEAL && opt_authorized_policy) {
-		if (!pcr_authorized_policy_seal_secret(tpm2key_fmt, opt_authorized_policy, opt_input, opt_output))
+		if (!pcr_authorized_policy_seal_secret(target, opt_authorized_policy, opt_input, opt_output))
 			return 1;
 
 		return 0;
 	}
 
 	if (action == ACTION_UNSEAL) {
-		if (tpm2key_fmt) {
-			/* input is the sealed secret, output is the cleartext */
-			if (!pcr_policy_unseal_tpm2key (opt_input, opt_output))
-				return 1;
-		} else
-		if (opt_rsa_public_key) {
-			/* input is the sealed secret, output is the cleartext */
-			if (!pcr_authorized_policy_unseal_secret(pcr_selection, opt_pcr_policy, opt_rsa_public_key, opt_input, opt_output))
-				return 1;
-		} else {
-			if (!pcr_unseal_secret(pcr_selection, opt_input, opt_output))
-				return 1;
-		}
+		if (!pcr_unseal_secret(target, pcr_selection, opt_pcr_policy, opt_rsa_public_key, opt_input, opt_output))
+			return 1;
 
 		return 0;
 	}
@@ -1240,8 +1310,12 @@ main(int argc, char **argv)
 	if (opt_stop_event && (!opt_from || strcmp(opt_from, "eventlog")))
 		usage(1, "--stop-event only makes sense when using event log");
 
+	/* If pcr_selection is NULL, the programmer must have been sloppy. */
+	if (pcr_selection == NULL)
+		fatal("BUG: action %u should have parsed a PCR selection argument", action);
+
 	pred = predictor_new(pcr_selection, opt_from, opt_eventlog_path,
-			opt_output_format);
+			opt_output_format, opt_boot_entry);
 
 	if (opt_stop_event)
 		predictor_set_stop_event(pred, opt_stop_event, !opt_stop_before);
@@ -1256,12 +1330,11 @@ main(int argc, char **argv)
 			predictor_report(pred);
 	} else
 	if (action == ACTION_SEAL) {
-		/* TBD - seal secret against a set of PCR values */
-		if (!pcr_seal_secret(tpm2key_fmt, &pred->prediction, opt_input, opt_output))
+		if (!pcr_seal_secret(target, &pred->prediction, opt_input, opt_output))
 			return 1;
 	} else
 	if (action == ACTION_SIGN) {
-		if (!pcr_policy_sign(tpm2key_fmt, &pred->prediction, opt_rsa_private_key, opt_input, opt_output, opt_policy_name))
+		if (!pcr_policy_sign(target, &pred->prediction, opt_rsa_private_key, opt_input, opt_output, opt_policy_name))
 			return 1;
 	}
 
